import java.lang.annotation.*;

@Retention(value = RetentionPolicy.CLASS)
@Target(value = {ElementType.CONSTRUCTOR,ElementType.METHOD})
public @interface GhostComponent
{
	/*
	Array of String storing the methods required in the body.
	They refer to the static type - that is the root of the subhierarchy.
	All of its subtypes will be then considered as possible receivers.
	These methods refer to classes belonging to the client program only.
	The name of classes has to be included in the class method signature, according to the full package class name, as well as arguments and return types.
	This is how to write them as strings: 
		RETURN_TYPE PACKAGE_CLASS_NAME.METHOD(ARG1,...,ARGN)
	E.g. = "java.lang.String package.class_name.toString(int[],java.lang.Object)"
	The class name of an inner class is return package.classname$inner_class_name.method(args)
	Constructor calls (that is, new ()) are annotated as <init>.
	If the array is empty, declare it as '{}' in the annotation, NOT as 'null'.

	*/
	String[] req_meths();
	/*
	Array of String indicating the name of the exceptions declared to be caught within the method.
	The name of the exception has to match with its namespace declaration (user-defined or not).
	E.g.: "java.lang.NullPointerException", "java.io.IOException","java.lang.ArithmeticException" etc...
	If the array is empty, declare it as '{}' in the annotation, NOT as 'null'.
	*/

	String[] handlers();
}
