(*
sawjaExtension -
This file includes the functions useful to parse and generate a callgraph from a BIR program
Authors: Pedro de Carvalho Gomes <pedrodcg@kth.se>
         Attilio Picoco <picoco@kth.se>
Date: 2013-02-02
*)

open Javalib_pack
open JBasics
open Javalib
open Sawja_pack
open JProgram
open JBir

let get_method_calls_bir p cs cm =
  let l = ref Ptmap.empty in
  let f_lookup = p.static_lookup_method in
    begin
      match cm with
	| {cm_implementation = Java code}
	    when
	      ClassMethodMap.mem cm.cm_class_method_signature p.parsed_methods ->
	    let ms = cm.cm_signature in
	      Array.iteri
		(fun pp op ->
		   match op with
		     | JBir.InvokeVirtual _ | JBir.InvokeStatic _ | JBir.InvokeNonVirtual _ | JBir.New _ ->
			 let lookup = (f_lookup cs ms pp) in
			   l := Ptmap.add pp lookup !l
		     | _ -> ())
		(Lazy.force code).code
	| _ -> ()
    end;
    !l

(* Taken from JProgram
type callgraph = ((class_name * method_signature * int)
		  * (class_name * method_signature)) list
*)

(**
* @param p JBir.t input program
* @return !calls JProgram related Sawja callgraph
*
* It gets a program in input and it gives the related Sawja callgraph in output *)
let get_callgraph_bir p =
  let methodcalls2callsite cs ms calls =
    let l = ref [] in
      Ptmap.iter
	(fun pp cmset ->
	   ClassMethodSet.iter
	     (fun ccms ->
		let (ccs,cms) = cms_split ccms in
		  l := ((cs,ms,pp),(ccs,cms)) :: !l
	     ) cmset
	) calls;
      !l in
  let calls = ref [] in
    iter
      (fun ioc ->
	 match ioc with
	   | Interface {i_info = {i_name = cs; i_initializer = Some cm}} ->
               calls :=
                 (methodcalls2callsite cs cm.cm_signature
		    (get_method_calls_bir p cs cm)) @ !calls
           | Interface _ -> ()
	   | Class c ->
	       MethodMap.iter
		 (fun _ m ->
		    match m with
		      | ConcreteMethod cm ->
			  let cs = c.c_info.c_name in
			    calls :=
			      (methodcalls2callsite cs cm.cm_signature
				 (get_method_calls_bir p cs cm))
			    @ !calls
		      | AbstractMethod _ -> ()
		 ) c.c_info.c_methods
      ) p;
    !calls

(*
(*
type 'a t = {cl:'a JProgram.node;
	       meth:'a Javalib_pack.Javalib.concrete_method;
	       pc:int;}
*)
type my_pp =
{
	class_node:'a JProgram.node; (* it can be both Interface or Class *)
	meth:'a Javalib_pack.Javalib.concrete_method;
	cms:JBasics.class_method_signature; (* complete class method signature *)
	pc:int;(* program counter - position in the array *)
	atomic_prop:string list (* Can either be an empty list or contain up to two atomic propositions: "ret" and one exception type *) (**MAYBE NO NEED FOR THIS HERE*)
};;

exception NoCode of (class_name * method_signature)

(*
let get_first_pp prog cs ms : 'a t =
	match get_node prog cs with
	| Interface {i_info = {i_initializer = Some m}} as c
	  when ms_equal m.cm_signature ms ->
	  {cl=c;meth=m;pc=0;}
	| Class c as cl' when MethodMap.mem ms c.c_info.c_methods ->
	  begin
	    match MethodMap.find ms c.c_info.c_methods with
	      | ConcreteMethod m->
		  {cl=cl'; meth=m; pc=0;}
	      | _ -> raise (NoCode (cs,ms))
	  end
	| _ -> raise (NoCode (cs,ms))
*)

(*
let get_node program cs =
  ClassMap.find cs program.classes <---- IT IS A MAP (KEY: class_name VALUE: class nodes)
*)
(** What if we do not have the classes or their implementation? *)
let get_first_pp prog cs ms : my_pp =
	match JProgram.get_node prog cs with
	| JProgram.Interface {i_info = {i_initializer = Some m}} as c
	  when JBasics.ms_equal m.cm_signature ms ->
		(***)
	  {class_node = c; meth = m; cms = (JBasics.make_cms cs ms); pc = 0; atomic_prop = []}
	| JProgram.Class c as cl' when JBasics.MethodMap.mem ms c.c_info.c_methods ->
	  begin
	    match JBasics.MethodMap.find ms c.c_info.c_methods with
	      | Javalib.ConcreteMethod m->
		(***)
		  {class_node = cl'; meth = m; cms = (JBasics.make_cms cs ms); pc = 0; atomic_prop = []}
	      | _ -> raise (NoCode (cs,ms))
	  end
	| _ -> raise (NoCode (cs,ms))
;;


(*
let get_meth (pp:'a t) : 'a concrete_method =
    pp.meth
*)
let get_meth (pp:my_pp) : 'a Javalib_pack.Javalib.concrete_method =
	pp.meth
;;

let get_cms (pp:my_pp) : JBasics.class_method_signature =
	pp.cms
;;
(*
let get_pc (pp:'a t) : int = pp.pc
*)
let get_pc (pp:my_pp) : int = pp.pc;;

(* To create cfg nodes *)
(*
let make_cfg_node ?(tag_ret=false) ?(tag_excp=false) ?(tag_label=[]) pp node_pc =
   {
      cmsig = (SawjaExtension.get_cms pp);
      pc = node_pc;
      nlabels = tag_label;
      excp = tag_excp;
      ret = tag_ret;
   };;

let make_cfg_node2 ?(tag_ret=false) ?(tag_excp=false) ?(tag_label=[]) pp =
   {
      cmsig = (SawjaExtension.get_cms pp);
      pc = (SawjaExtension.get_pc pp);
      nlabels = tag_label;
      excp = tag_excp;
      ret = tag_ret;
   };;

(* creates a cfg node *)
and cfg_node ?(tag_ret=false) ?(tag_excp=false) ?(tag_label=[]) pp =
(* Create the node *)
 { 
    cmsig = curr_meth_cms ;
    pc = (SawjaExtension.get_pc pp);
    (* Add entry label only if PC is zero *)
    nlabels = if ( ((SawjaExtension.get_pc pp) = 0) && (not tag_ret) && (not tag_excp)) then ([ent_lbl]@tag_label) else tag_label;
    excp = tag_excp; 
    ret = tag_ret;
 };
*)
(*
  let normal_successors pp =
    match get_opcode pp with
      | OpIf (_,l)
      | OpIfCmp (_,l) ->
	  [next_instruction pp;goto_relative pp l]
      | OpJsr l
      | OpGoto l -> [goto_relative pp l]
      | OpRet _ -> (* all instruction following a jsr are returned. *)
	  let code = get_code pp in
	  let i = ref 0 in
	  let l = ref [] in
	    while !i < Array.length code do
	      begin
		match code.(!i) with
		  | OpJsr _ ->
		      l := next_instruction (goto_absolute pp !i)::!l
		  | _ -> ()
	      end;
	      incr i;
	    done;
	    !l
      | OpTableSwitch (default,_,_,others) ->
	  Array.fold_left
	    (fun ppl jmp -> goto_relative pp jmp::ppl)
	    [goto_relative pp default]
	    others
      | OpLookupSwitch (default,others) ->
	  List.fold_left
	    (fun ppl (_,jmp) -> goto_relative pp jmp::ppl)
	    [goto_relative pp default]
	    others
      | OpThrow
      | OpReturn _ -> []
      | OpInvalid
      | OpBreakpoint ->
	  raise (Class_structure_error "Instructions Invalid and Breakpoint are not authorized")
      | _ -> [next_instruction pp]
*)

  let goto_absolute pp i : 'a t = {pp with pc=i;} (* -> : my_pp *)

  let goto_relative pp jmp : 'a t ={pp with pc=pp.pc+jmp;} (* -> : my_pp *)

  let get_code pp : jopcodes =
    match pp.meth.cm_implementation with (** What if we do not have the code??? *)
      | Java c -> (Lazy.force c).c_code
      | Native -> raise (NoCode (get_name pp.cl,pp.meth.cm_signature))

  let get_opcode pp : jopcode = 
    (get_code pp).(pp.pc)
      
  let next_instruction pp =
    let opcodes = get_code pp
    and i = ref (succ pp.pc)
    in
      while opcodes.(!i) = OpInvalid do
	Pervasives.incr i;
      done;
      goto_absolute pp !i

  let normal_successors (pp:my_pp) =
    match get_opcode pp with
      | OpIf (_,l)
      | OpIfCmp (_,l) ->
	  [next_instruction pp;goto_relative pp l]
      | OpJsr l
      | OpGoto l -> [goto_relative pp l]
      | OpRet _ -> (* all instruction following a jsr are returned. *)
	  let code = get_code pp in
	  let i = ref 0 in
	  let l = ref [] in
	    while !i < Array.length code do
	      begin
		match code.(!i) with
		  | OpJsr _ ->
		      l := next_instruction (goto_absolute pp !i)::!l
		  | _ -> ()
	      end;
	      incr i;
	    done;
	    !l
      | OpTableSwitch (default,_,_,others) ->
	  Array.fold_left
	    (fun ppl jmp -> goto_relative pp jmp::ppl)
	    [goto_relative pp default]
	    others
      | OpLookupSwitch (default,others) ->
	  List.fold_left
	    (fun ppl (_,jmp) -> goto_relative pp jmp::ppl)
	    [goto_relative pp default]
	    others
      | OpThrow
      | OpReturn _ -> []
      | OpInvalid
      | OpBreakpoint ->
	  raise (Class_structure_error "Instructions Invalid and Breakpoint are not authorized")
      | _ -> [next_instruction pp]
*)
